@@ -2,7 +2,7 @@
 #include "mapper.h"
 #include "servicedef.h"
 
-#include <google/protobuf/dynamic_message.h>
+#include "pb/gpb_mapping.h"
 
 #include <sstream>
 
@@ -20,6 +20,23 @@ using namespace upb::googlepb;
     #define newXS(a, b, c) Perl_newXS(aTHX_ const_cast<char *>(a), b, const_cast<char *>(c))
 #endif
 
+namespace {
+    SV *get_stack_trace(pTHX) {
+        dSP;
+
+        PUSHMARK(SP);
+        PUTBACK;
+
+        call_pv("Google::ProtocolBuffers::Dynamic::_stack_trace", G_SCALAR);
+
+        SPAGAIN;
+        SV *stack_trace = POPs;
+        PUTBACK;
+
+        return stack_trace;
+    }
+}
+
 void Dynamic::CollectErrors::AddError(const string &filename, int line, int column, const string &message) {
     croak("Error during protobuf parsing: %s:%d:%d: %s", filename.c_str(), line, column, message.c_str());
 }
@@ -37,8 +54,11 @@ MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
         no_redefine_perl_names(false),
         accessor_style(GetAndSet),
         client_services(Disable),
-        numeric_bool(false),
+        boolean_style(Perl),
+        default_decoder(Upb),
         fail_ref_coercion(false) {
+    stack_trace = get_stack_trace(aTHX);
+
     if (options_ref == NULL || !SvOK(options_ref))
         return;
     if (!SvROK(options_ref) || SvTYPE(SvRV(options_ref)) != SVt_PVHV)
@@ -64,44 +84,48 @@ MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
 
 #undef BOOLEAN_OPTION
 
-    if (SV **value = hv_fetchs(options, "accessor_style", 0)) {
-        const char *buf = SvPV_nolen(*value);
+#define START_STRING_VALUE(name) \
+    if (SV **value = hv_fetchs(options, #name, 0)) { \
+        const char *buf = SvPV_nolen(*value); \
+        \
+        if (0 == 1)
 
-        if (strEQ(buf, "get_and_set"))
-            accessor_style = GetAndSet;
-        else if (strEQ(buf, "plain_and_set"))
-            accessor_style = PlainAndSet;
-        else if (strEQ(buf, "single_accessor"))
-            accessor_style = SingleAccessor;
-        else if (strEQ(buf, "plain"))
-            accessor_style = Plain;
-        else
-            croak("Invalid value '%s' for 'accessor_style' option", buf);
+#define STRING_VALUE(field, string, value) \
+        else if (strEQ(buf, #string)) \
+            field = value
+
+#define END_STRING_VALUE(name) \
+        else \
+            croak("Invalid value '%s' for '" #name "' option", buf); \
     }
 
-    if (SV **value = hv_fetchs(options, "client_services", 0)) {
-        const char *buf = SvPV_nolen(*value);
+    START_STRING_VALUE(default_decoder);
+    STRING_VALUE(default_decoder, upb, Upb);
+    STRING_VALUE(default_decoder, bbpb, Bbpb);
+    END_STRING_VALUE(default_decoder);
 
-        if (strEQ(buf, "disable"))
-            client_services = Disable;
-        else if (strEQ(buf, "noop"))
-            client_services = Noop;
-        else if (strEQ(buf, "grpc_xs"))
-            client_services = GrpcXS;
-        else
-            croak("Invalid value '%s' for 'client_services' option", buf);
-    }
+    START_STRING_VALUE(accessor_style);
+    STRING_VALUE(accessor_style, get_and_set, GetAndSet);
+    STRING_VALUE(accessor_style, plain_and_set, PlainAndSet);
+    STRING_VALUE(accessor_style, single_accessor, SingleAccessor);
+    STRING_VALUE(accessor_style, plain, Plain);
+    END_STRING_VALUE(accessor_style);
 
-    if (SV **value = hv_fetchs(options, "boolean_values", 0)) {
-        const char *buf = SvPV_nolen(*value);
+    START_STRING_VALUE(client_services);
+    STRING_VALUE(client_services, disable, Disable);
+    STRING_VALUE(client_services, noop, Noop);
+    STRING_VALUE(client_services, grpc_xs, GrpcXS);
+    END_STRING_VALUE(client_services);
 
-        if (strEQ(buf, "perl"))
-            numeric_bool = false;
-        else if (strEQ(buf, "numeric"))
-            numeric_bool = true;
-        else
-            croak("Invalid value '%s' for 'boolean_values' option", buf);
-    }
+    START_STRING_VALUE(boolean_values);
+    STRING_VALUE(boolean_style, perl, Perl);
+    STRING_VALUE(boolean_style, numeric, Numeric);
+    STRING_VALUE(boolean_style, json, JSON);
+    END_STRING_VALUE(boolean_values);
+
+#undef START_STRING_VALUE
+#undef STRING_VALUE
+#undef END_STRING_VALUE
 }
 
 Dynamic::Dynamic(const string &root_directory) :
@@ -427,29 +451,51 @@ void Dynamic::map_message_prefix_recursive(pTHX_ const Descriptor *descriptor, c
 }
 
 void Dynamic::check_package(pTHX_ const string &perl_package, const string &pb_name) {
-    if (used_packages.find(perl_package) == used_packages.end())
+    string marker_name = perl_package + "::mapped_from";
+    SV *marker_sv = get_sv(marker_name.c_str(), 0);
+
+    if (!marker_sv || !SvOK(marker_sv))
         return;
 
-    croak("Package '%s' has already been used in a mapping", perl_package.c_str());
+    croak("Package '%s' has already been mapped from %s",
+          perl_package.c_str(), SvPV_nolen(marker_sv));
+}
+
+void Dynamic::mark_package(pTHX_ const string &perl_package, SV *stack_trace) {
+    string marker_name = perl_package + "::mapped_from";
+#ifdef GV_ADDMULTI
+    const int get_sv_flags = GV_ADDMULTI;
+#else
+    const int get_sv_flags = GV_ADD;
+
+    // this is just to avoid the 'used only once' warning
+    get_sv(marker_name.c_str(), get_sv_flags);
+#endif
+
+    SV *marker_sv = get_sv(marker_name.c_str(), get_sv_flags);
+
+    sv_setsv(marker_sv, stack_trace);
 }
 
 void Dynamic::map_message(pTHX_ const Descriptor *descriptor, const string &perl_package, const MappingOptions &options) {
-    check_package(aTHX_ perl_package, descriptor->full_name());
+    bool define_perl_names = !options.no_redefine_perl_names || gv_stashpvn(perl_package.data(), perl_package.size(), 0) == NULL;
+    if (define_perl_names)
+        check_package(aTHX_ perl_package, descriptor->full_name());
     if (descriptor_map.find(descriptor->full_name()) != descriptor_map.end())
         croak("Message '%s' has already been mapped", descriptor->full_name().c_str());
     if (options.use_bigints)
         load_module(PERL_LOADMOD_NOIMPORT, newSVpvs("Math::BigInt"), NULL);
-    bool define_perl_names = !options.no_redefine_perl_names || gv_stashpvn(perl_package.data(), perl_package.size(), 0) == NULL;
     HV *stash = gv_stashpvn(perl_package.data(), perl_package.size(), GV_ADD);
     const MessageDef *message_def = def_builder.GetMessageDef(descriptor);
+    const gpd::pb::Descriptor *gpd_descriptor = gpd::pb::map_pb_descriptor(&descriptor_set, descriptor, descriptor_loader.pool());
     if (is_map_entry(message_def, options.implicit_maps))
         // it's likely I will regret this const_cast<>
         upb_msgdef_setmapentry(const_cast<MessageDef *>(message_def), true);
-    Mapper *mapper = new Mapper(aTHX_ this, message_def, stash, options);
+    Mapper *mapper = new Mapper(aTHX_ this, message_def, gpd_descriptor, stash, options);
 
     // the map owns the reference from Mapper constructor, and is unreffed in ~Dynamic
     descriptor_map[message_def->full_name()] = mapper;
-    used_packages.insert(perl_package);
+    mark_package(aTHX_ perl_package, options.stack_trace);
     pending.push_back(mapper);
 
     if (define_perl_names)
@@ -472,7 +518,13 @@ void Dynamic::bind_message(pTHX_ const string &perl_package, Mapper *mapper, HV
     }
 
     copy_and_bind(aTHX_ "set_decoder_options", perl_package, mapper);
-    copy_and_bind(aTHX_ "decode", perl_package, mapper);
+    copy_and_bind(aTHX_ "decode_upb", perl_package, mapper);
+    copy_and_bind(aTHX_ "decode_bbpb", perl_package, mapper);
+    if (options.default_decoder == MappingOptions::Upb) {
+        copy_and_bind(aTHX_ "decode_upb", "decode", perl_package, mapper);
+    } else {
+        copy_and_bind(aTHX_ "decode_bbpb", "decode", perl_package, mapper);
+    }
     copy_and_bind(aTHX_ "encode", perl_package, mapper);
     copy_and_bind(aTHX_ "decode_json", perl_package, mapper);
     copy_and_bind(aTHX_ "encode_json", perl_package, mapper);
@@ -578,7 +630,7 @@ void Dynamic::map_enum(pTHX_ const EnumDescriptor *descriptor, const string &per
     EnumMapper *mapper = new EnumMapper(aTHX_ this, enum_def);
 
     mapped_enums.insert(descriptor->full_name());
-    used_packages.insert(perl_package);
+    mark_package(aTHX_ perl_package, options.stack_trace);
 
     HV *stash = gv_stashpvn(perl_package.data(), perl_package.size(), GV_ADD);
 
@@ -601,7 +653,7 @@ void Dynamic::map_service(pTHX_ const ServiceDescriptor *descriptor, const strin
         croak("Service '%s' has already been mapped", descriptor->full_name().c_str());
 
     mapped_services.insert(descriptor->full_name());
-    used_packages.insert(perl_package);
+    mark_package(aTHX_ perl_package, options.stack_trace);
 
     ServiceDef *service_def = new ServiceDef(descriptor->full_name());
 
