@@ -7,6 +7,8 @@
 
 #include "upb/bridge.h"
 
+#include "pb/decoder.h"
+
 #include "descriptorloader.h"
 #include "sourcetree.h"
 #include "ref.h"
@@ -23,6 +25,12 @@ class MethodMapper;
 class ServiceDef;
 
 struct MappingOptions {
+    enum BoolStyle {
+        Perl = 1,
+        Numeric = 2,
+        JSON = 3,
+    };
+
     enum AccessorStyle {
         GetAndSet = 1,
         PlainAndSet = 2,
@@ -36,6 +44,11 @@ struct MappingOptions {
         GrpcXS = 2,
     };
 
+    enum Decoder {
+        Upb = 0,
+        Bbpb = 1,
+    };
+
     bool use_bigints;
     bool check_required_fields;
     bool explicit_defaults;
@@ -46,10 +59,12 @@ struct MappingOptions {
     bool implicit_maps;
     bool decode_blessed;
     bool fail_ref_coercion;
-    bool numeric_bool;
     bool no_redefine_perl_names;
+    BoolStyle boolean_style;
     AccessorStyle accessor_style;
     ClientService client_services;
+    Decoder default_decoder;
+    SV *stack_trace;
 
     MappingOptions(pTHX_ SV *options_ref);
 };
@@ -98,6 +113,7 @@ private:
     void map_service_noop(pTHX_ const google::protobuf::ServiceDescriptor *descriptor, const std::string &perl_package, const MappingOptions &options, ServiceDef *service_def);
     void map_service_grpc_xs(pTHX_ const google::protobuf::ServiceDescriptor *descriptor, const std::string &perl_package, const MappingOptions &options, ServiceDef *service_def);
     void check_package(pTHX_ const std::string &perl_package, const std::string &pb_name);
+    void mark_package(pTHX_ const std::string &perl_package, SV *stack_trace);
     std::string pbname_to_package(pTHX_ const std::string &pb_name, const std::string &perl_package_prefix);
 
     OverlaySourceTree overlay_source_tree;
@@ -107,12 +123,12 @@ private:
     upb::googlepb::DefBuilder def_builder;
     CollectErrors die_on_error;
     STD_TR1::unordered_map<std::string, const Mapper *> descriptor_map;
-    STD_TR1::unordered_set<std::string> used_packages;
     STD_TR1::unordered_set<std::string> mapped_enums;
     STD_TR1::unordered_set<std::string> mapped_services;
     STD_TR1::unordered_set<const google::protobuf::FileDescriptor *> files;
     std::vector<Mapper *> pending;
     std::vector<MethodMapper *> pending_methods;
+    gpd::pb::DescriptorSet descriptor_set;
 };
 
 // equivalent to FieldDescriptor::has_presence() in src/google/protobuf/descriptor.h
