@@ -150,21 +150,23 @@ SV *Mapper::DecoderHandlers::get_and_mortalize_target() {
 
 namespace {
 #if PERL_VERSION < 18
-    inline SSize_t GPD_av_top_index(pTHX_ AV *av) {
-        return AvFILL(av);
-    }
-    #define av_top_index(av) GPD_av_top_index(aTHX_ (av))
-
     #define READ_XDIGIT(s)  ((0xf & (isDIGIT(*(s))     \
                                     ? (*(s)++)         \
                                     : (*(s)++ + 9))))
 #endif
 
-#if PERL_VERSION < 14
-    inline void GPD_warn_sv(pTHX_ SV *mess) {
-        warn("%" SVf, mess);
+#if PERL_VERSION < 12
+    bool GPD_is_invariant_string(const U8 *s, STRLEN len) {
+        const U8 *end = s + (len ? len : strlen((const char *) s));
+
+        for (; s < end; ++s) {
+            if (!UTF8_IS_INVARIANT(*s))
+                return false;
+        }
+
+        return true;
     }
-    #define warn_sv(sv) GPD_warn_sv(aTHX_ (sv))
+    #define is_invariant_string(s, len) GPD_is_invariant_string(s, len)
 #endif
 
     inline void set_perl_bool(pTHX_ SV *target, bool value) {
@@ -263,6 +265,7 @@ size_t Mapper::DecoderHandlers::on_append_string(DecoderHandlers *cxt, const int
     THX_DECLARE_AND_GET;
 
     STRLEN cur = SvCUR(cxt->string);
+    SvGROW(cxt->string, cur + len); // necessary for JSON, where the size hint is 0
     memcpy(SvPVX(cxt->string) + cur, buf, len);
     SvCUR_set(cxt->string, cur + len);
 
